--!parent
--!encoding=utf-8

--****************************************************************************
-- @file lib "oauth"
-- @author Valentin Schmidt
-- @version 0.9
-- @requires lib "curl" (+xtra "Curl")
--****************************************************************************

property $

property _verbose
property _user_agent

property _oauth_version
property _oauth_signature_method

property _oauth_consumer_key
property _oauth_consumer_secret

property _oauth_request_token
property _oauth_request_token_secret

property _oauth_access_token
property _oauth_access_token_secret

property _oauth_verifier

property _base64_lookup

property _bitshift_list
property _upper_masks_list

----------------------------------------
-- @constructor
----------------------------------------
on new (me)

	-- libs
	$.import("curl")

	me._oauth_version = "1.0"
	me._oauth_signature_method = "HMAC-SHA1"

	me._verbose = FALSE
	me._user_agent = VOID

	-- used for Base64 calculation
	me._base64_lookup = [65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,80,81,82,83,84,\
	85,86,87,88,89,90,97,98,99,100,101,102,103,104,105,106,107,108,109,110,111,112,113,\
	114,115,116,117,118,119,120,121,122,48,49,50,51,52,53,54,55,56,57,43,47]

	-- used for SHA-1 calculation
	me._bitshift_list = [\
	1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024, 2048, 4096, 8192, 16384, 32768, 65536, 131072, 262144, 524288,\
	1048576, 2097152, 4194304, 8388608, 16777216, 33554432, 67108864, 134217728, 268435456, 536870912,\
	1073741824, the maxinteger+1]
	me._upper_masks_list = [2147483647, 1073741823, 536870911, 268435455, 134217727, 67108863, 33554431, 16777215,\
	8388607, 4194303, 2097151, 1048575, 524287, 262143, 131071, 65535, 32767, 16383, 8191, 4095, 2047, 1023, 511,\
	255, 127, 63, 31, 15, 7, 3, 1, 0]

	return me
end

----------------------------------------
-- Toggles if curl prints connection details to stdout.
-- @param {bool} flag
----------------------------------------
on set_verbose (me, flag)
	me._verbose = flag
end

----------------------------------------
-- Set User-Agent used for all requests
-- @param {string} user_agent
----------------------------------------
on set_user_agent (me, user_agent)
	me._user_agent = user_agent
end

----------------------------------------
-- Set OAuth consumer key
-- @param {string} consumer_key
-- @param {string} consumer_secret
----------------------------------------
on set_consumer_key (me, consumer_key, consumer_secret)
	me._oauth_consumer_key = consumer_key
	me._oauth_consumer_secret = consumer_secret
end

----------------------------------------
-- Set OAuth request token
-- @param {string} request_token
-- @param {string} request_token_secret
----------------------------------------
on set_request_token (me, request_token, request_token_secret)
	me._oauth_request_token = request_token
	me._oauth_request_token_secret = request_token_secret
end

----------------------------------------
-- Set OAuth verifier
-- @param {string} verifier
----------------------------------------
on set_verifier (me, verifier)
	me._oauth_verifier = verifier
end

----------------------------------------
-- Set OAuth access token
-- @param {string} access_token
-- @param {string} access_token_secret
----------------------------------------
on set_access_token (me, access_token, access_token_secret)
	me._oauth_access_token = access_token
	me._oauth_access_token_secret = access_token_secret
end

----------------------------------------
-- @return {string[]} Array containg string oauth_request_token and oauth_request_token_secret
----------------------------------------
on get_request_token (me)
	return [me._oauth_request_token, me._oauth_request_token_secret]
end

----------------------------------------
-- @return {string[]} Array containg string oauth_access_token and oauth_access_token_secret
----------------------------------------
on get_access_token (me)
	return [me._oauth_access_token, me._oauth_access_token_secret]
end

----------------------------------------
-- Get OAuth request token
-- @param {string} base_url E.g. "http://www.flickr.com/services/oauth/request_token"
-- @param {string} oauth_callback_name
-- @param {list} [err] Optional, receives error string
-- @return {bool} success
----------------------------------------
on query_request_token (me, base_url, oauth_callback_name, err)
	if voidP(err) then
		err = []
	end if
	key = me._oauth_consumer_secret&"&"
	oauth_nonce = me._create_nonce(32)
	oauth_timestamp = me._get_timestamp()
	query_str = ""
	if not voidP(oauth_callback_name) then
		put "oauth_callback="&me._urlencode(oauth_callback_name)&"&" after query_str
	end if
	put "oauth_consumer_key="&me._oauth_consumer_key after query_str
	put "&oauth_nonce="&oauth_nonce after query_str
	put "&oauth_signature_method="&me._oauth_signature_method after query_str
	put "&oauth_timestamp="&oauth_timestamp after query_str
	put "&oauth_version=1.0" after query_str

	data = "GET&"&me._urlencode(base_url)&"&"&me._urlencode(query_str)
	sig = me._b64_hmac_sha1(key, data)
	url = base_url&"?"&query_str&"&oauth_signature="&me._urlencode(sig)

	response_code = []
	txt = me._http_query(url, VOID, response_code, "get")

	if (response_code[1]<>200) THEN
		err[1] = "Error getting request token! Network error: "&response_code
		return FALSE
	else
		if ilk(txt)=#bytearray then
			txt.position = 1
			txt = txt.readRawString(txt.length)
		end if
		res = me._parse_str(txt)
		if (res["oauth_callback_confirmed"]="TRUE") then
			me._oauth_request_token = res["oauth_token"]
			me._oauth_request_token_secret = res["oauth_token_secret"]
			return TRUE
		else
			err[1] = "Error getting request token!"
			return FALSE
		end if
	end if
end

----------------------------------------
--
----------------------------------------
on authorize (me, perms)
	u = "https://www.flickr.com/services/oauth/authorize?oauth_token="&me._oauth_request_token
	if not voidP(perms) then
		put "&perms="&perms after u
	end if
	-- open in default browser
	--browserName("C:\Program Files\Mozilla Firefox\Firefox.exe")
	gotoNetPage(u)
end

----------------------------------------
-- Get OAuth access token
-- @param {string} base_url E.g. "http://www.flickr.com/services/oauth/access_token"
-- @param {list} [err] Optional, receives error string
-- @return {bool} success
----------------------------------------
on query_access_token (me, base_url, err)
	if voidP(err) then
		err = []
	end if

	key = me._oauth_consumer_secret&"&"&me._oauth_request_token_secret
	oauth_nonce = me._create_nonce(32)
	oauth_timestamp = me._get_timestamp()

	query_str = ""
	put "oauth_consumer_key="&me._oauth_consumer_key after query_str
	put "&oauth_nonce="&oauth_nonce after query_str
	put "&oauth_signature_method="&me._oauth_signature_method after query_str
	put "&oauth_timestamp="&oauth_timestamp after query_str
	put "&oauth_token="&me._oauth_request_token after query_str
	put "&oauth_verifier="&me._oauth_verifier after query_str
	put "&oauth_version="&me._oauth_version after query_str

	data = "GET&"&me._urlencode(base_url)&"&"&me._urlencode(query_str)
	sig = me._b64_hmac_sha1(key, data)
	url = base_url&"?"&query_str&"&oauth_signature="&me._urlencode(sig)

	response_code = []
	txt = me._http_query(url, VOID, response_code, "get")

	if (response_code[1]<>200) then
		err[1] = "Error getting access token! Network error: "&response_code
		return FALSE
	else
		if ilk(txt)=#bytearray then
			txt.position = 1
			txt = txt.readRawString(txt.length)
		end if
		res = me._parse_str(txt)
		if (not voidP(res["oauth_token"]) AND not voidP(res["oauth_token_secret"])) then
			me._oauth_access_token = res["oauth_token"]
			me._oauth_access_token_secret = res["oauth_token_secret"]
			return TRUE
		else
			err[1] = "Error getting access token!"
			return FALSE
		end if
	end if
end

----------------------------------------
-- Make OAuth (REST) API call
-- @param {string} base_url E.g. "http://www.flickr.com/services/oauth/access_token"
-- @param {propList} [param_obj] Optional, data to pass to url
-- @param {list} [response_code] Optional, receives HTTP response_code
-- @param {string} [method="get"]
-- @return {bytearray|integer} result
----------------------------------------
on call_api (me, base_url, param_obj, response_code, method)
	if ilk(param_obj)<>#propList then
		param_obj = [:]
	end if
	data = me.build_query(param_obj)
	sig = me.get_signature(base_url, data, method)
	put "&oauth_signature="&me._urlencode(sig) after data
	if method="post" then
		res = me._http_query(base_url, data, response_code, "post")
	else
		put base_url&"?" before data
		res = me._http_query(data, VOID, response_code, "get")
	end if
	return res
end

----------------------------------------
-- Make asynchronous OAuth (REST) API call
-- @param {string} base_url E.g. "http://www.flickr.com/services/oauth/access_token"
-- @param {propList} [param_obj=VOID] Optional, data to pass to url
-- @param {string} [method="get"]
-- @param {symbol} result_callback
-- @param {object} [result_target=_movie]
-- @param {propList} [custom_data=VOID]
----------------------------------------
on call_api_async (me, base_url, param_obj, method, result_callback, result_target, custom_data)
	if ilk(param_obj)<>#propList then
		param_obj = [:]
	end if
	data = me.build_query(param_obj)
	sig = me.get_signature(base_url, data, method)
	put "&oauth_signature="&me._urlencode(sig) after data
	if method="post" then
		me._http_query(base_url, data, VOID, "post", result_callback, result_target, custom_data)
	else
		put base_url&"?" before data
		me._http_query(data, VOID, VOID, "get", result_callback, result_target, custom_data)
	end if
end

----------------------------------------
-- @param {propList} param_obj
-- @return {string}
----------------------------------------
on build_query (me, param_obj)
	oauth_nonce = me._create_nonce(32)
	oauth_timestamp = string(me._get_timestamp())
	if ilk(param_obj)<>#propList then
		param_obj = [:]
	end if
	param_obj["oauth_consumer_key"] = me._oauth_consumer_key
	param_obj["oauth_nonce"] = oauth_nonce
	param_obj["oauth_signature_method"] = me._oauth_signature_method
	param_obj["oauth_timestamp"] = oauth_timestamp
	param_obj["oauth_token"] = me._oauth_access_token
	param_obj["oauth_version"] = me._oauth_version
	sort(param_obj) -- sort props
	return $.curl.httpBuildQuery(param_obj)
end

----------------------------------------
-- @param {string} base_url
-- @param {string} query_str
-- @param {string} [method="get"]
-- @return {string}
----------------------------------------
on get_signature (me, base_url, query_str, method)
	if method="POST" then
		method="POST" -- force upper case
	else
		method="GET"
	end if
	key = me._oauth_consumer_secret&"&"&me._oauth_access_token_secret
	data = method&"&"&me._urlencode(base_url)&"&"&me._urlencode(query_str)
	return me._b64_hmac_sha1(key, data)
end

--**************************************
-- PRIVATE
--**************************************

----------------------------------------
-- HTTP/HTTPS request (GET or POST)
-- @param {string} url
-- @param {string} [data=VOID]
-- @param {list} [response_code] Optional, receives integer response_code
-- @param {string} [method="get"] Optional, either "get"(=default) or "post"
-- @param {symbol} result_callback
-- @param {object} [result_target=_movie]
-- @param {any} [custom_data=VOID]
-- @return {bytearray|integer|VOID} result
----------------------------------------
on _http_query (me, url, data, response_code, method, result_callback, result_target, custom_data)
	ch = $.curl.init()
	if me._verbose then
		ch.setOption($.curl.CURLOPT.VERBOSE, 1)
	end if
	ch.setOption($.curl.CURLOPT.URL, url)
	ch.setOption($.curl.CURLOPT.SSL_VERIFYPEER, 0)
	if not voidP(me._user_agent) then
		ch.setOption($.curl.CURLOPT.USERAGENT, me._user_agent)
	end if
	if stringP(data) then
		if method="post" then
			ch.setOption($.curl.CURLOPT.POSTFIELDS, data)
		else
			put "?"&data after url
		end if
	end if
	if symbolP(result_callback) then
		$.curl.execAsyncDetached(ch, result_callback, result_target, 1, custom_data)
	else
		res = ch.exec(1)
		if listP(response_code) then
			response_code[1] = integer(ch.getInfo($.curl.CURLINFO.RESPONSE_CODE))
		end if
		ch.close()
		return res
	end if
end

----------------------------------------
-- Creates a "nonce" random string of specified length
-- @param {integer} len
-- @return {string}
----------------------------------------
on _create_nonce (me, len)
 	res = ""
	str = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
	rmax = length(str)
	repeat with i = 1 to len
		put str.char[random(rmax)] after res
	end repeat
	return res
end

----------------------------------------
-- @param {string|bytearray} key
-- @param {string|bytearray} data
-- @return {string}
----------------------------------------
on _b64_hmac_sha1 (me, key, data)
	if stringP(key) then
		key = bytearray(key)
	end if
	if stringP(data) then
		data = bytearray(data)
	end if
	blocksize = 64
	if (key.length>blocksize) then
		key = me._sha1(key)
	end if
	if (key.length<blocksize) then
		repeat with i = key.length+1 to blocksize
			key[i] = 0
		end repeat
	end if
	ipad = byteArray(blocksize)
	repeat with i = 1 to blocksize
		ipad[i] = 54 -- 0x36
	end repeat
	opad = byteArray(blocksize)
	repeat with i = 1 to blocksize
		opad[i] = 92 -- 0x5c
	end repeat

	ba = me._baXor(key,ipad)
	ba.position = ba.length+1
	ba.writeByteArray(data)
	hmac = me._sha1(ba)

	ba = me._baXor(key,opad)
	ba.position = ba.length+1
	ba.writeByteArray(hmac)
	hmac = me._sha1(ba)

	return me._base64_encode(hmac)
end

----------------------------------------
-- bitXor on bytearrays
-- @param {bytearray} a
-- @param {bytearray} b
-- @return {bytearray}
----------------------------------------
on _baXor (me, a, b)
	len = min(a.length,b.length)
	ba = byteArray()
	repeat with i = 1 to len
		ba[i] = bitXor(a[i], b[i])
	end repeat
	return ba
end

----------------------------------------
-- Returns timestamp for current date/time
-- @return {integer}
----------------------------------------
on _get_timestamp (me)
	date_obj = the systemdate
	ts = (date_obj - date(1970,1,1))*86400 + date_obj.seconds
	return ts
end

----------------------------------------
-- URL encodes a string (as specified in RFC 3986)
-- @param {string} str
-- @return {string}
----------------------------------------
on _urlencode (me, str)
	allowed = "abcdefghijklmnopqrstuvwxyz0123456789-_."
	res = ""
	len = str.length
	repeat with i = 1 to len
		c = str.char[i]
		if not (allowed contains c) then
			c = me._hexencode(c)
		end if
		put c after res
	end repeat
	return res
end

----------------------------------------
-- URL decodes a string
-- @param {string} str
-- @return {string}
----------------------------------------
on _urldecode (me, str)
	res = ""
	ba = bytearray()
	len = str.length
	repeat with i = 1 to len
		c = str.char[i]
		if (c = "%") then
			ba.writeInt8(rgb(str.char[i+1..i+2]).blue)
			i = i + 2
		else if (c = "+") then
			ba.writeInt8(32)
		else
			ba.writeInt8(chartonum(c))
		end if
	end repeat
	ba.position = 1
	return ba.readRawString(ba.length)
end

----------------------------------------
-- Parses string like a query string, returns propList
-- @param {string} str
-- @return {propList}
----------------------------------------
on _parse_str (me, str)
	res = [:]
	parts = me._explode("&", str)
	repeat with p in parts
		tmp = me._explode("=", p)
		if count(tmp)>1 then
			res.addProp(tmp[1], me._urldecode(tmp[2]))
		else
			res[tmp[1]] = ""
		end if
	end repeat
	return res
end

----------------------------------------
-- ByteArray -> ASCII String
-- @param {string|bytearray} ba_data
-- @return {string}
----------------------------------------
on _base64_encode (me, ba_data)
	if stringP(ba_data) then
		ba_data = byteArray(ba_data)
	end if
	ba_out = byteArray()
	i=1
	strRemainder = ba_data.length mod 3
	strLength = ba_data.length - strRemainder
	repeat while (i < strLength)
		a = ba_data[i]
		b = ba_data[i+1]
		c = ba_data[i+2]
		byte1 = me._base64_lookup[ ( bitAnd( a, 252 ) / 4 ) + 1]
		byte2 = me._base64_lookup[ ( bitAnd( (a*256)+b, 1008 ) / 16 ) + 1]
		byte3 = me._base64_lookup[ ( bitAnd( (b*256)+c, 4032 ) / 64 ) + 1]
		byte4 = me._base64_lookup[ ( bitAnd( c, 63 ) ) + 1]
		ba_out.writeInt32( byte1 + byte2*256 + byte3*65536 + byte4*16777216)
		i = i + 3
	end repeat
	if (strRemainder = 1) then
		a = ba_data[i]
		ba_out.writeInt8( me._base64_lookup[ ( bitAnd( a, 252 ) / 4 ) + 1] )
		ba_out.writeInt8( me._base64_lookup[ ( bitAnd( a*256, 1008 ) / 16 ) + 1] )
		ba_out.writeRawString( "==", 2)
	else if (strRemainder = 2) then
		a = ba_data[i]
		b = ba_data[i+1]
		ba_out.writeInt8( me._base64_lookup[ ( bitAnd( a, 252 ) / 4 ) + 1] )
		ba_out.writeInt8( me._base64_lookup[ ( bitAnd( (a*256)+b, 1008 ) / 16 ) + 1] )
		ba_out.writeInt8( me._base64_lookup[ ( bitAnd( b*256, 4032 ) / 64 ) + 1] )
		ba_out.writeRawString( "=", 1)
	end if
	ba_out.position = 1
	return ba_out.readRawString(ba_out.length)
end

----------------------------------------
-- Calculates SHA-1 hash of string or bytearray
-- @param {bytearray|string} input
-- @return {string} (20 bytes)
----------------------------------------
on _sha1 (me, input)
	if stringP(input) then
		input = bytearray(input)
	end if
	balen = input.length
	bitlen = balen * 8
	endian_org = input.endian
	input.endian = #bigEndian
	-- padding
	pad1_pos = bitlen/32*4 +1
	pad2_pos = ((((bitlen + 64)/512) * 16) + 15)*4 +1

	w = [] -- list of integers
	w[80]=0

	a =	1732584193
	b = -271733879
	c = -1732584194
	d =	271733878
	e = -1009589776

	cntx = (((bitlen + 64)/512) * 16) + 15 +1
	cntk = cntx/16-1
	repeat with k = 0 to cntk
		i = k*16

		olda = a
		oldb = b
		oldc = c
		oldd = d
		olde = e

		repeat with j = 0 to 79
			if(j < 16) then
				p = (i+j)*4 +1
				if p=pad1_pos then
					v = 0
					if pad1_pos	<=balen then
						v=v+input[pad1_pos]*16777216
					end if
					if pad1_pos+1<=balen then
						v=v+input[pad1_pos+1]*65536
					end if
					if pad1_pos+2<=balen then
						v=v+input[pad1_pos+2]*256
					end if
					if pad1_pos+3<=balen then
						v=v+input[pad1_pos+3]
					end if
					w[j +1] = bitOr( v, me._bitshift(128, -(24 - bitlen mod 32)) )
				else if p=pad2_pos then
					w[j +1] = bitlen
				else if p>balen then
					w[j +1] = 0
				else
					input.position = p
					w[j +1] = input.readInt32()
				end if
			else
				w[j +1] = me._rol( bitXOr(bitXOr(bitXOr(w[j-2], w[j-7]), w[j-13]), w[j-15]), 1)
			end if

			if (j < 20) then
				sha1_ft = bitOr( bitAnd(b, c), bitAnd(bitNot(b), d))
				sha1_kt = 1518500249
			else if (j < 40) then
				sha1_ft = bitXor(bitXor(b, c), d)
				sha1_kt = 1859775393
			else if (j < 60) then
				sha1_ft = bitOr( bitOr(bitAnd(b, c), bitAnd(b, d)), bitAnd(c, d))
				sha1_kt = -1894007588
			else
				sha1_ft = bitXor(bitXor(b, c), d)
				sha1_kt = -899497514
			end if

			t = me._rol(a, 5) + sha1_ft + e + w[j +1] + sha1_kt

			e = d
			d = c
			c = me._rol(b, 30)
			b = a
			a = t

		end repeat

		a = a + olda
		b = b + oldb
		c = c + oldc
		d = d + oldd
		e = e + olde

	end repeat

	input.endian = endian_org -- reset, just in case a littleEndian bytearray was passed

	ba = bytearray()
	ba.endian = #bigEndian
	ba.writeInt32(a)
	ba.writeInt32(b)
	ba.writeInt32(c)
	ba.writeInt32(d)
	ba.writeInt32(e)
	ba.position = 1
	return ba
	--return ba.readRawString(ba.length)
end

----------------------------------------
-- Bitwise rotate a 32-bit number to the left.
-- analog to JS: return (num << cnt) | (num >>> (32 - cnt));
-- @param {integer} num
-- @param {integer} cnt
-- @return {integer}
----------------------------------------
on _rol (me, num, cnt)
	if num=0 then
		return 0
	end if
	return bitOr(me._bitshift(num, -cnt), me._bitshift(num, (32 - cnt)))
end

----------------------------------------
-- arg_amount<0 => bitShiftleft
-- arg_amount>0 => bitShiftRightZeroFill (>>> in JS)
-- @param {integer} arg_val
-- @param {integer} arg_amount - max. supported value: 31
-- @return {integer}
----------------------------------------
on _bitshift (me, arg_val, arg_amount)
	if arg_amount=0 then
		return arg_val
	end if
	if (arg_amount > 0) then
		if (arg_val > 0) then arg_val = arg_val / me._bitshift_list[arg_amount+1]
		else --means highest bit is set - remove it!
			arg_val = bitAnd(arg_val, 2147483647)
			arg_val = arg_val / me._bitshift_list[arg_amount+1]
			arg_val = arg_val + me._bitshift_list[32-arg_amount]
		end if
	else
		arg_val = bitAnd(me._upper_masks_list[-arg_amount], arg_val)
		arg_val = arg_val * me._bitshift_list[-arg_amount+1]
	end if
	return arg_val
end

----------------------------------------
-- Returns sequence of hex codes (prepended with "%") for specified UTF-8 string
-- @param {string|bytearray} ba
-- @return {string}
----------------------------------------
on _hexencode (me, ba)
	res = ""
	if stringP(ba) then
		ba = bytearray(ba)
	end if
	str = me._upper(ba.toHexString(1,ba.length))
	cnt = str.word.count
	repeat with i = 1 to cnt
		put "%" & str.word[i] after res
	end repeat
	return res
end

----------------------------------------
-- Explode string into list based on single char delimiter
-- @param {string} delim
-- @param {string} str
-- @return {list}
----------------------------------------
on _explode (me, delim, str)
	l = []
	if voidP(str) then
		return l
	end if
	od = _player.itemDelimiter
	_player.itemDelimiter = delim
	cnt = str.item.count
	repeat with i = 1 to cnt
		l.add(str.item[i])
	end repeat
	_player.itemDelimiter = od
	return l
end

----------------------------------------
-- Lower to upper case (ASCII only)
-- @param {string} str
-- @return {string}
----------------------------------------
on _upper (me, input)
	alphabet = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
	len = input.length
	out = ""
	repeat with i = 1 to len
		c = input.char[i]
		pos = offset(c, alphabet)
		if pos > 0 then
			put alphabet.char[pos] after out
		else
			put c after out
		end if
	end repeat
	return out
end
